#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `posh` package."""
import pytest
import os

from bs4 import BeautifulSoup, FeatureNotFound

from collections import OrderedDict
import datetime
from dateutil.relativedelta import relativedelta

from random import random
from requests import get

from posh import account
from posh.product_search import ProductSearch
from posh.product import Product, get_past_date

#from sqlalchemy import create_engine
#from sqlalchemy.orm import scoped_session, sessionmaker

product_search = ProductSearch()


def test_build_request():
    possible_arguments = OrderedDict({
        'brand': 'LuLaRoe',
        'sex': 'Women',
        'category': 'Dresses',
        'subcategory': 'Mini',
        'color': 'Black',
        'size': 'M',
        'sort': 'added_desc',
        'type': 'closet',
        'price': '26-50'
    })

    possible_arguments = dict(sorted(possible_arguments.items(),
                                     key=lambda x: random()))

    # Make sure that arguments are ordered correctly
    # regardless of supplied order.

    for i in range(1, 5):
        possible_arguments = dict(sorted(possible_arguments.items(),
                                         key=lambda x: random()))
        test_string = product_search._build_request(possible_arguments)
        assert test_string == 'https://poshmark.com/brand/LuLaRoe-Wom' + \
            'en-Dresses-Mini-color-Black?sort_by=added_des' + \
            'c&size%5B%5D=M&condition=closet&price%5B%5D=26-50'


def test_execute_search():
    possible_arguments = OrderedDict({
        'brand': 'LuLaRoe',
        'sex': 'Women',
        'category': 'Dresses',
        'subcategory': 'Mini',
        'color': 'Black',
        'size': 'M',
        'sort': 'added_desc',
        'type': 'closet',
        'price': '26-50'
    })

    for i in range(1, 5):
        possible_arguments = dict(sorted(possible_arguments.items(),
                                         key=lambda x: random()))
        product_search.execute_search(possible_arguments, page_number=1)
        assert len(product_search.results) > 10
        product_search.results = []
    product_search.execute_search(possible_arguments, page_number=47)
    assert len(product_search.results) == 0


def test_build_product_from_url():
    product = Product(url='https://poshmark.com/listing/Lularo' +
                      'e-Carly-5c2d86fcbaebf68a9b6893b0')
    product._build_product_from_url(product_search.session)
    assert product.owner == 'cmunger81'
    assert product.brand == 'LuLaRoe'
    assert product.price == 25.0
    assert product.size == ['S']


def test_search_multiple_pages():
    product_search.search_multiple_pages(50, arguments={
        'brand': 'Vera Bradley',
        'size': 'OS',
        'sex': 'Women',
        'category': 'Bags',
        'subcategory': 'Laptop Bags',
        'color': 'Pink'
    })
    assert len(product_search.results) >= 545
    # More products are added all the time so it'd be tricky to pin
    # down an exact number that wouldn't require changing all the time.
    # However, the search should always return >545 unique items.
    # May need to be changed in the future if Vera Bradley goes out
    # of style. (possibly occurred already?)


def test_brand_search():
    product_search.results = []
    product_search.search_multiple_pages(1, arguments={
        'brand': 'rag & bone'
        })
    for result in product_search.results:
        assert result.brand.lower() == 'rag & bone'


def test_product_update():
    possible_arguments = OrderedDict({
        'brand': 'LuLaRoe',
        'sex': 'Women',
        'category': 'Dresses',
        'subcategory': 'Mini',
        'color': 'Black',
        'size': 'M',
        'sort': 'added_desc',
        'type': 'closet',
        'price': '26-50'
    })
    product_search.execute_search(possible_arguments)

    first_result = product_search.results[0]
    assert first_result.updated_at is None
    first_result.update(product_search.session, built_from='tile')
    assert isinstance(first_result.updated_at, datetime.datetime)


def test_get_past_date():
    test_strings = ['Updated 3 minutes ago', 'Updated 3 hours ago',
                    'Updated an hour ago', 'Updated 3 days ago',
                    'Updated Yesterday']

    for string in test_strings:
        assert type(get_past_date(string)) == datetime.datetime

    with pytest.raises(ValueError):
        string = "Updated 90 eons ago."
        get_past_date(string)

    assert get_past_date('Updated 3 hours ago').hour == \
        (datetime.datetime.today() - relativedelta(hours=3)).hour

    assert type(get_past_date('Updated 9/27/2018')) == datetime.datetime


def test_prepare_for_db_insert():
    arguments = OrderedDict({
        'brand': 'LuLaRoe',
        'sex': 'Women',
        'category': 'Dresses',
        'subcategory': 'Mini',
        'color': 'Black',
        'size': 'M',
        'sort': 'added_desc',
        'type': 'closet',
        'price': '26-50'
    })

    product_search.execute_search(arguments)

    first_result = product_search.results[1]

    assert first_result.description is None

    first_result.update(product_search.session)

    assert isinstance(first_result.description, str)


@pytest.mark.skip(reason="Needs to be customized to each user's DB instance.")
def test_insert_into_db():
    engine = create_engine(
        "postgres:///nick:nickspassword@localhost/test:5432")

    connection = engine.connect()

    db_session = scoped_session(sessionmaker(
        autocommit=True, autoflush=True, bind=engine))

    product = Product(url='https://poshmark.com/listi' +
                      'ng/Lularoe-Carly-5c2d86fcbaebf68a9b6893b0')
    product._build_product_from_url(product_search.session)

    product.insert_into_db(db_session)


def test_search_by_query():
    product_search.results = []

    arguments = OrderedDict({
        'query': 'signed jersey'
    })

    product_search.execute_search(arguments)
    #  It should be made clear somewhere that this method searches
    #  for items exactly the same as if the search occurred on Poshmark.com
    #  - which means that all sorts of non-jersey products get included
    #  in this search.  Consider adding a filter arg to include
    #  only items with query in title, etc. instead of trusting
    #  Poshmark's search.

    result = any([i.title.lower() for i in product_search.results
                  if 'jersey' in i.title.lower()])
    assert result


def test_category_search():
    product_search.results = []

    arguments = {'category': 'Makeup',
                 'sex': 'Women'}

    request_str = product_search._build_request(arguments)

    r = get(request_str, headers={'User-Agent': 'Posh'})

    try:
        soup = BeautifulSoup(r.content, 'lxml')
    except FeatureNotFound:
        soup = BeautifulSoup(r.content, 'html.parser')

    assert soup.find_all(
        'span', attrs={'itemprop': 'name'})[1].text == 'Makeup'

    # The second span with name itemprop is the currently selected category.


def test_get_pictures():
    product = Product(url='https://poshmark.com/listing/Lularo' +
                      'e-Carly-5c2d86fcbaebf68a9b6893b0')
    product._build_product_from_url(product_search.session)
    assert len(product.pictures) == 1


def test_get_images():
    product = Product(url='https://poshmark.com/listing/Lularo' +
                      'e-Carly-5c2d86fcbaebf68a9b6893b0')
    product._build_product_from_url(product_search.session)
    product.get_images(folder_path=os.curdir)

    assert 'Lularoe Carly-5c2d86fcbaebf68a9b6893' + \
        'b0-cmunger81_0.jpg' in product.images[0]


def test_strict_search():
    product_search.results = []
    product_search.search_multiple_pages(pages=2, arguments={
        'query': 'signed jersey'
    }, strict=True)

    assert all(['signed jersey' in i.title.lower()
                for i in product_search.results])


def test_search_over_time():
    product_search.results = []
    product_search.search_product_price_over_time(arguments=OrderedDict({
        'query': 'NWT vera wang'}
    ), strict=False)
    assert len(product_search.results) >= 500


@pytest.mark.skip(reason="Need to find a way to actually test this.")
def test_plot_time_price_tuples():
    product_search.results = []
    product_search.search_product_price_over_time(arguments=OrderedDict({
        'query': 'NWT vera wang'}
    ), strict=False)
    product_search.plot_time_price_tuples()

@pytest.mark.skip(reason="Not properly implemented yet.")
def test_account_login():
    test_account = Account(username='ntucker12312', password='testing_for_posh')
    assert not test_account.check_login()
    
    test_account.login()
    assert test_account.check_login()

@pytest.mark.skip(reason="Not implemented yet.")
def test_product_like():
    test_account = Account(username='ntucker12312', password='testing_for_posh')
    test_account.login()

    product = Product(url='https://poshmark.com/listing/Lularo' +
                      'e-Carly-5c2d86fcbaebf68a9b6893b0')
    product._build_product_from_url(product_search.session)

    product.like(test_account)




from bs4 import BeautifulSoup, FeatureNotFound
import datetime
from dateutil.parser import parse as date_parser
from dateutil.relativedelta import relativedelta
import os


def get_past_date(str_days_ago):
    """
    Converts arbitrary "updated at" strings to proper datetimes.
    """

    str_days_ago = str_days_ago.replace(' an ', ' 1 ')

    #  When it's been <2 hours, Poshmark returns "an hour ago"
    #  instead of "1 hour ago" - which, without this replacement,
    #  screws up the later date parsing.

    today = datetime.datetime.today()
    split_str = str_days_ago.split()

    if 'Yesterday' in split_str:
        return datetime.datetime.now() - relativedelta(days=1)

    elif len(split_str) < 2:   # Could all of this be replaced by date_parser?
        date = date_parser(str_days_ago[8:])
        return date

    elif len(split_str) == 2:
        date = date_parser(str_days_ago[8:])
        return date

    elif len(split_str) > 2:
        if 'minute' in split_str[2]:
            return datetime.datetime.now() - \
                datetime.timedelta(minutes=int(split_str[1]))

        elif 'hour' in split_str[2]:
            date = datetime.datetime.now() - \
                relativedelta(hours=int(split_str[1]))
            return date

        elif 'day' in split_str[2]:
            date = today - relativedelta(days=int(split_str[1]))
            return date

        else:
            return date_parser(str_days_ago[8:])

    else:
        raise ValueError(
            f'Supplied date str is {split_str}, ' +
            'which doesn\'t match any supported formats.')


class Product:
    """
    An individual product on Poshmark. Can currently be built
    from either a search or built with a URL directly.
    Some attributes can only be gathered via search, while some can
    only be gathered via URL.  A product's full information can be
    gathered when the product originates from a search, but not
    when the product originates from a supplied URL.
    """

    def __init__(self, url, posted_at=None, owner=None,
                 brand=None, price=None, size=None, listing_id=None,
                 title=None, pictures=None, updated_at=None,
                 description=None, colors=None, comments=None):
        self.__soup = None

        self._built_from = None

        self.images = None

        self.url = url
        self.posted_at = posted_at
        self.owner = owner
        self.brand = brand
        self.price = price
        self.size = size
        self.listing_id = listing_id
        self.title = title
        self.pictures = pictures
        self.updated_at = updated_at

        # The following attributes are only available if built from URL.

        self.description = description
        self.colors = colors
        self.comments = comments

    def insert_into_db(self, db_session, table_name='product'):
        self.update(self.session)
        query = f""" \
                INSERT INTO {table_name} (url, owner, brand, price,\
                size, listing_id, title, pictures, description,\
                colors, comments, built_from)
                VALUES
                (\
                '{self.url}', '{self.owner}','{self.brand}',
                 {self.price}, '{', '.join([i for i in self.size])}',
                 '{self.listing_id}',
                 '{self.title}', '{self.pictures}',
                 '{self.description}', '{self.colors}', '{self.comments}',
                 '{self._built_from}')
                """
        db_session.execute(query)
        return

    def _build_product_from_tile(self, tile, session):
        """
        Builds products from tiles, i.e. returned search results.
        """

        self.session = session  # This is lazy
        self._built_from = 'tile'

        self.posted_at = tile['data-created-at']
        self.owner = tile['data-creator-handle']
        self.brand = tile.get('data-post-brand')  # Not always there.
        self.price = tile['data-post-price']
        self.size = tile['data-post-size']
        self.listing_id = tile['id']
        self.title = tile.find('a')['title']

    def _build_product_from_url(self, session):
        self.session = session  # This is lazy

        if not self._built_from:
            self._built_from = 'url'

        try:
            soup = BeautifulSoup(session.get(self.url).content, 'lxml')
        except FeatureNotFound:
            soup = BeautifulSoup(session.get(self.url).content, 'html.parser')

        self.__soup = soup
        self._get_pictures()

        self.owner = soup.find('div', class_='handle').text[1:]
        self.brand = soup.find('meta', attrs={'property': 'product:brand'}
                           ).get('content')
        self.price = float(soup.find('meta',
                                     attrs={'property': 'product:price:amount'}
                                     ).get('content'))
        self.size = [i.text.strip() for i in
                     soup.find('div', class_='size-con').find_all('label')]
        self.listing_id = self.url.split('-')[-1]
        self.title = soup.find('h1', class_='title').text
        updated_at = soup.find('span', class_='time').text
        self.updated_at = get_past_date(updated_at)
        self.description = soup.find('div', class_='description').text
        self.colors = None  # Not yet implemented.
        self.comments = soup.find(
            'div', class_='comments')  # Not yet fully implemented.

    def update(self, session, built_from='tile'):
        if built_from == 'tile':
            self._build_product_from_url(session)

    def _get_pictures(self):
        if not self.__soup:
            try:
                self.__soup = BeautifulSoup(
                    self.session.get(self.url).content, 'lxml')
            except FeatureNotFound:
                self.__soup = BeautifulSoup(
                    self.session.get(self.url).content, 'html.parser')

        pictures = self.__soup.find_all('img', attrs={'itemprop': 'image'})
        picture_urls = [i.get('data-img-src') for i in pictures if i.get(
            'data-img-src')]
        self.pictures = picture_urls
        return

    def get_images(self, folder_path=None):
        self.images = []
        if len(self.pictures) == 0:
            self._get_pictures()

        for index, link in enumerate(self.pictures):
            r = self.session.get(link)
            file_name = \
                f'{self.title}-{self.listing_id}-{self.owner}_{index}.jpg'
            if folder_path:
                if not os.path.isdir(folder_path):
                    os.mkdir(folder_path)
                file_name = f'{folder_path}/{file_name}'
            with open(file_name, 'wb') as f:
                f.write(r.content)
                self.images.append(file_name)




from bs4 import BeautifulSoup, FeatureNotFound
import requests


class Account:
	def __init__(self, username, password, session=requests.Session()):
		self.username = username
		self.password = password
		self.session = session

	def login(self):
		r = self.session.get('https://poshmark.com/login')
		try:
			soup = BeautifulSoup(r.content, 'lxml')
		except FeatureNotFound:
			soup = BeautifulSoup(r.content)

		authenticity_token = soup.find('input', attrs={'name': 'authenticity_token'}).get('value')

		r = self.session.post(
			'https://poshmark.com/login',
			params={
			'utf8': '✓',
			'authenticity_token': authenticity_token,
			'login_form[iobb]': None,
			'login_form[username_email]': self.username,
			'login_form[password]': self.password
			})

	def check_login(self):
		r = self.session.get('https://poshmark.com/feed?login=true')
		if self.username in str(r.content):
			return True
		else:
			return False


from bs4 import BeautifulSoup, FeatureNotFound
from collections import OrderedDict
import datetime
from dateutil.relativedelta import relativedelta
from posh.product import Product
import operator
import requests
import time
import sys


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]


def datestring_to_timestamp(str):
    return time.mktime(time.strptime(str, "%Y-%m-%d %H:%M:%S %z"))


def timestamp_to_datestring(timestamp):
    return time.strftime('%Y-%m-%d %H:%M:%S %z', time.localtime(timestamp))


class ProductSearch:
    """
    May change substantially in future versions. Currently,
    handles one search and it's results.
    """

    def __init__(self):
        self.session = requests.Session()
        self.set_headers()
        self.results = []
        self.time_price_tuples = [] # Used in search_over_time method.

    def _format_argument(self, argument, value, addition):
        if argument == 'subcategory':
            addition = addition.replace(' ', '_')
        elif argument == 'brand':
            addition = addition.replace(' ', '_')
        elif argument == 'query':
            addition = addition.replace(' ', '+')
        return addition

    def _build_request(self, arguments: dict):
        """
        May change substantially in future versions.
        Currently creates the string for a search given
        a dict with at least one possible_argument.
        """

        string = 'https://poshmark.com/'
        possible_arguments = OrderedDict(
            {'query': 'search?query=',
             'brand': 'brand/',
             'sex': '-',
             'category': '-',
             'subcategory': '-',
             'color': '-color-',
             'sort': '?sort_by=',
             'size': '&size%5B%5D=',
             'type': '&condition=',
             'price': '&price%5B%5D='
             }
        )
        for argument, value in possible_arguments.items():
            if argument == 'sort' and not arguments.get(argument):
                string += possible_arguments['sort'] + 'added_desc'
            elif arguments.get(argument) and argument == 'query':
                addition = possible_arguments[argument] + arguments[argument]
                addition = self._format_argument(argument, value, addition)
                string += addition
            elif arguments.get(argument):
                addition = possible_arguments[argument] + arguments[argument]
                addition = self._format_argument(argument, value, addition)
                string += addition

        if string[21] == '-':
            if 'category' in arguments.keys():
                string = string[:21] + 'category/' + string[22:]

        if 'query' in string:
            string = string.replace('?sort_by', '&sort_by')

        if 'max_id' in arguments.keys():
            if 'search?query' in string:
                string = string.replace(
                    'search?query',
                    f'search?max_id={arguments["max_id"]}&query')
            else:
                string += f'&max_id={arguments["max_id"]}"'

        #  Too many statements to accomplish proper string mashing.
        #  Find a way to tighten the above.

        return string

    def _check_strictness(self, tile, arguments):
        """
        'Strictness' refers to whether or not the search result *must* contain
        the string searched for in the title - results will be returned if it's
        contained within the description as well.
        """
        title = tile.find('a').get('title').lower()
        for key, value in arguments.items():
            if str(value).lower() not in title:
                return False
            else:
                continue
        return True

    def set_headers(self, headers=None):
        """
        If provided, sets headers. Otherwise,
        sets User-Agent to "Posh"
        """

        if not headers:
            headers = {'User-Agent': 'Posh'}

        self.session.headers.update(headers)

        return

    def execute_search(self, arguments, page_number=None,
                       items=None, strict=False):
        """
        Given a request_str, executes the associated search.
        Gathers results, turns their HTML into Product objects,
        then adds them to the ProductSearch object's results attr.
        """
        if page_number:
            if page_number == 1:
                pass
            else:
                arguments.update({'max_id': int(page_number)})

        request_str = self._build_request(arguments)

        r = self.session.get(request_str)

        try:
            soup = BeautifulSoup(r.content, 'lxml')
        except FeatureNotFound:
            soup = BeautifulSoup(r.content, 'html.parser')

        tiles = soup.find_all('div', class_='tile')

        for tile in tiles:
            strictness_pass = self._check_strictness(tile, arguments)
            if strict and strictness_pass:
                #  There needs to be a better way to do this.
                p = Product(
                    url=f"https://poshmark.com{tile.find('a').get('href')}")
                p._build_product_from_tile(tile, self.session)
                self.results.append(p)
                continue
            elif strict and not strictness_pass:
                continue
            elif not strict:
                p = Product(
                    url=f"https://poshmark.com{tile.find('a').get('href')}")
                p._build_product_from_tile(tile, self.session)
                self.results.append(p)
                continue
        return

    def search_multiple_pages(self, pages, arguments, strict=False):
        for page in range(1, pages + 1):
            old_results_len = len(self.results)
            self.execute_search(arguments=arguments, page_number=page,
                                items=self.results, strict=strict)
            new_results_len = len(self.results)
            if not strict:
                # Strictness check can cause real new results to not be caught.
                if new_results_len == old_results_len:
                    return

    def search_product_price_over_time(self, arguments, strict=False):
        self.search_multiple_pages(
            arguments=arguments, pages=48, strict=strict)
        time_sorted_products = sorted(
            self.results, key=operator.attrgetter('posted_at'))

        for product_chunk in chunks(time_sorted_products, 25):
            avg_time = float(sum([
                datestring_to_timestamp(i.posted_at)
                for i in product_chunk])) / max(len(product_chunk), 1)
            avg_price = sum(
                [float(i.price.replace('$', '').replace(',', ''))
                 for i in product_chunk])
            self.time_price_tuples.append(
                (timestamp_to_datestring(avg_time), avg_price)
            )

    def plot_time_price_tuples(self, arguments=None, strict=False):
        if 'matplotlib' not in sys.modules.keys():
            import matplotlib.pyplot as plt
        if len(self.time_price_tuples) == 0:
            self.search_product_price_over_time(arguments, strict)
        plt.ylabel('Time')
        plt.xlabel('Price')
        plt.plot(
            [i[0] for i in self.time_price_tuples],
            [i[1] for i in self.time_price_tuples]  # I don't like this
        )
        plt.show()
